#include "dpm.hpp"

DPM::DPM( double _threshhold, string model):
    threshold(_threshhold) 
{
    detector = DPMDetector::create(vector<string>{model});
}


void DPM::get_detections(string& image_path, int frame, vector<vector<float>>& sort_detections, vector<vector<double>>& results){

    cv::Mat image = cv::imread(image_path, cv::IMREAD_UNCHANGED);

    if(!image.data) {
        cout << endl << "Failed to load image at " << image_path << endl << "Ensure correct file path" << endl;
        exit(1);
    }
    // Apply OpenCV's DPM object detection algo on this image
    vector<DPMDetector::ObjectDetection> ds;
    detector->detect(image, ds);

    for (auto& det: ds) {
        if(det.score > threshold){ 
            sort_detections.emplace_back( vector<float>{(float)det.rect.x , (float)det.rect.y, (float)(det.rect.width + det.rect.x), 
                                                (float)(det.rect.height + det.rect.y), (float)det.score} );
        	
            results.emplace_back( 
                vector<double>{ (double)(frame + 1), -1.0,
                    (double)det.rect.x , 
                    (double)det.rect.y, 
                    (double)det.rect.width, 
                    (double)det.rect.height, 
                    (double)det.score, -1.0, -1.0, -1.0}   
            );
		}
	}
}