#include "evalTrack.hpp"


void evalTrack(string& seq, VVVD& pre_proc_tracks, trackingResults& results) {
            
	VVVD tracks(pre_proc_tracks.size());
	VVVD rawGT(pre_proc_tracks.size());

	ifstream file("data/" + seq + "/gt/gt.txt");
	size_t frame_shift = 0;
	loadtxt_by_frame(rawGT, file, frame_shift);

	double vis_thresh = -1.0; // no minimum thresh for cleaning tracking data 
	cleanData(rawGT, pre_proc_tracks, tracks, vis_thresh);
	// assert same number of frames
	assert(rawGT.size() == tracks.size());

	VVVD gt(tracks.size());
	cleanGT(rawGT, gt, vis_thresh);

	VVD mot_track;
	VVD mot_gt;
	formatToLongList(gt, mot_gt);
	formatToLongList(tracks, mot_track);

	clear_mot_hun(mot_gt, mot_track, results);
	ID_measures(mot_gt, mot_track, results);

	printTrackResults(results, seq);
    //computeResults(results);
}

void ID_measures(VVD& rawGT, VVD& rawTrack, trackingResults& res){
	
	const int INF = 1e8;
	size_t total_gt = rawGT.size();
	size_t total_tracks = rawTrack.size();
	size_t uniqGT = (size_t)rawGT.back()[1];
	size_t uniqTrack = (size_t)rawTrack.back()[1];

	VVVD gt(uniqGT);
	VVVD track(uniqTrack);

	formatByID(rawGT, rawTrack, gt, track);

	VVD costBlock(uniqGT, VD(uniqTrack));
	VVD fpBlock(uniqGT, VD(uniqTrack));
	VVD fnBlock(uniqGT, VD(uniqTrack));

	calcCostBlock(gt, track, costBlock, fpBlock, fnBlock);

	size_t dim = uniqGT + uniqTrack;
	VVD cost(dim, VD(dim, 0));
	for(size_t i = uniqGT; i < dim; ++i)
		for(size_t j = 0; j < uniqTrack; ++j)
			cost[i][j] = INF;
	for(size_t i = 0; i < uniqGT; ++i)
		for(size_t j = uniqTrack; j < dim; ++j)
			cost[i][j] = INF;

	VVD fp(dim, VD(dim, 0));
	VVD fn(dim, VD(dim, 0));

	for(size_t row = 0; row < costBlock.size(); ++row) {
		for(size_t col = 0; col < costBlock[row].size(); ++col){
			cost[row][col] = costBlock[row][col];
			fp[row][col] = fpBlock[row][col];
			fn[row][col] = fnBlock[row][col];
		}
	}
	// compute FP block		
	for(size_t i = 0; i < track.size(); ++i){
		cost[i + gt.size()][i] = track[i].size();
		fp[i + gt.size()][i] = track[i].size();
	}
	// compute FN block		
	for(size_t i = 0; i < gt.size(); ++i){
		cost[i][i + track.size()] = gt[i].size();
		fn[i][i + track.size()] = gt[i].size();
	}
	VI assignment(dim, -1);
	MinCostAssignment(cost, assignment);
	
	double IDFP = 0.0;
	double IDFN = 0.0;
	for(size_t i = 0; i < assignment.size(); ++i) {
    	if(assignment[i] != -1) {
			IDFP = IDFP + fp[i][assignment[i]];
			IDFN = IDFN + fn[i][assignment[i]];
		}
	}
	double IDTP = total_gt - IDFN;
	//% Sanity check
	assert((size_t)IDTP == (total_tracks - (size_t)IDFP));

	res.IDFP = IDFP;
 	res.IDFN = IDFN;
	res.IDTP = IDTP;
	res.total_tracks = (double)total_tracks;

	res.IDPrecision = (IDTP / (IDTP + IDFP) * 100);
	res.IDRecall = (IDTP / (IDTP + IDFN) * 100);
	res.IDF1 = (2*IDTP/((double)total_gt + (double)total_tracks) * 100);
}


/**
 * 
 * Following from Clear_MOT_Hun_mex
 */


void clear_mot_hun(VVD& gtMat, VVD& resMat, trackingResults& res)
{
	const double threshold = 0.5;
	const bool VERBOSE = false;

    int rowsGT = gtMat.size();  
	int rowsST = resMat.size(); 
    int Fgt=0, Ngt = 0, Nst = 0;

	VI unique_gt;  // counts number of unique gt ids at ith frame 
	vector<set<int>> framesPerGT(Ngt); // counts number of frames assiociated with ith gt ID

	// finds largest gt frame + ID
    for (int i = 0; i < rowsGT; i++) {
        int frame = gtMat[i][0];
        int ID = gtMat[i][1];
        if (frame > Fgt) Fgt = frame;
        if (ID > Ngt) Ngt = ID;

		while( (int)unique_gt.size() < ID + 1)
			unique_gt.emplace_back(0);
		unique_gt[ID] = 1;

		while( (int)framesPerGT.size() < ID + 1)
			framesPerGT.emplace_back();
		framesPerGT[ID].insert(frame);
    }

	// finds largest res frame + ID
    for (int i = 0; i < rowsST; i++) {
		int frame = resMat[i][0];
		if (frame > Fgt) Fgt = frame;
		int ID = resMat[i][1];
        if (ID > Nst) Nst = ID;
    } 

	double INF = 1e9;
	vector<unordered_map<int, int>> gtInd(Fgt); // Per Frame mapping gt ID to row index
	vector<unordered_map<int, int>> stInd(Fgt); // Per Frame mapping track ID to row index
	vector<unordered_map<int,int>> M(Fgt); // Holds per frame of gt row index to best track row index 
	VD mme(Fgt, 0); // ID Switchtes(mismatches)
	VD c(Fgt, 0); // matches found
	VD fp(Fgt, 0); // false positives
	VD m(Fgt, 0); // misses = false negatives
	VD g(Fgt, 0); // gt count for each frame
	vector<vector<double>> d(Fgt, vector<double>(Ngt, 0)); // all distances mapped to [0..1]
	vector<vector<int>> allfalsepos(Fgt, vector<int>(Nst, 0));

	// normalizing for 0 indexing
	for (int i = 0; i < rowsGT; i++) {
		int frame = gtMat[i][0] - 1; 
		int ID = gtMat[i][1] - 1; 
		gtInd[frame][ID] = i;
	}
	// normalizing for 0 indexing
	for (int i = 0; i < rowsST; i++) {
		int frame = resMat[i][0] - 1; 
		int ID = resMat[i][1] - 1; 
		stInd[frame][ID] = i;
	}

	// counts all gt per frame by ID
	for (int i = 0; i < Fgt; i++) {
		for (unordered_map<int, int>::iterator it = gtInd[i].begin(); it != gtInd[i].end(); it++) g[i]++;
	}
	
	for (int t = 0; t < Fgt; t++)
	{
		if (t > 0)
		{
			vector<int> mappings;
			for (unordered_map<int,int>::iterator it = M[t - 1].begin(); it != M[t - 1].end(); it++) mappings.push_back(it->first);
			std::sort(mappings.begin(), mappings.end());
			for (size_t k = 0; k < mappings.size(); k++)
			{
				unordered_map<int, int>::const_iterator foundGtind = gtInd[t].find(mappings[k]);
				unordered_map<int, int>::const_iterator foundStind = stInd[t].find(M[t - 1][mappings[k]]);

				if (foundGtind != gtInd[t].end() && foundStind != stInd[t].end())
				{
					bool matched = false;
				
					double gtleft, gttop, gtwidth, gtheight, stleft, sttop, stwidth, stheight;
					int rowgt = gtInd[t][mappings[k]];
					int rowres = stInd[t][M[t - 1][mappings[k]]];
					gtleft = gtMat[rowgt][2];//[index(rowgt, 2, rowsGT)];
					gttop = gtMat[rowgt][3];//[index(rowgt, 3, rowsGT)];
					gtwidth = gtMat[rowgt][4];//[index(rowgt, 4, rowsGT)];
					gtheight = gtMat[rowgt][5];//[index(rowgt, 5, rowsGT)];
					stleft = resMat[rowres][2];//[index(rowres, 2, rowsST)];
					sttop = resMat[rowres][3];//[index(rowres, 3, rowsST)];
					stwidth = resMat[rowres][4];//[index(rowres, 4, rowsST)];
					stheight = resMat[rowres][5];//[index(rowres, 5, rowsST)];
					double dist = 1 - adjusted_boxiou(gtleft, gttop, gtwidth, gtheight, stleft, sttop, stwidth, stheight);
					matched = (dist <= threshold);
					
					if (matched) {
						M[t][mappings[k]] = M[t - 1][mappings[k]];
						if (VERBOSE) printf("%d: preserve %d\n", t+1, mappings[k]+1);
					}
				}
			}
		}
		vector<int> unmappedGt, unmappedEs, stindt, findm;
		for (unordered_map<int,int>::iterator it = gtInd[t].begin(); it != gtInd[t].end(); it++) {
			unordered_map<int, int>::const_iterator found = M[t].find(it->first);
			if (found==M[t].end()) unmappedGt.push_back(it->first);
		}
		for (unordered_map<int,int>::iterator it = M[t].begin(); it != M[t].end(); it++) findm.push_back(it->second);
		for (unordered_map<int,int>::iterator it = stInd[t].begin(); it != stInd[t].end(); it++) stindt.push_back(it->first);

		sort(stindt.begin(), stindt.end());
		sort(findm.begin(), findm.end());
		set_difference(stindt.begin(), stindt.end(), findm.begin(), findm.end(), inserter(unmappedEs, unmappedEs.end()));

        sort(unmappedGt.begin(), unmappedGt.end());
        
		int squareSize = max(unmappedGt.size(), unmappedEs.size());
		vector<vector<double>> alldist(squareSize, vector<double>(squareSize, INF));

		if (VERBOSE)
		{
			printf("%d: UnmappedGTs: ", t+1);
			for (size_t i = 0; i < unmappedGt.size(); i++) printf("%d, ", unmappedGt[i]+1);
			printf("\n%d: UnmappedEs: ", t+1);
			for (size_t i = 0; i < unmappedEs.size(); i++) printf("%d, ", unmappedEs[i]+1);
			printf("\n");
		}

        int uid = 0; // Unique identifier
		for (size_t i = 0; i < unmappedGt.size(); i++)
		{
			for (size_t j = 0; j < unmappedEs.size(); j++)
			{
				int o = unmappedGt[i];
				int e = unmappedEs[j];

				double gtleft, gttop, gtwidth, gtheight, stleft, sttop, stwidth, stheight;
				int rowgt = gtInd[t][o];
				int rowres = stInd[t][e];
				gtleft = gtMat[rowgt][2];//[index(rowgt, 2, rowsGT)];
				gttop = gtMat[rowgt][3];//[index(rowgt, 3, rowsGT)];
				gtwidth = gtMat[rowgt][4];//[index(rowgt, 4, rowsGT)];
				gtheight = gtMat[rowgt][5];//[index(rowgt, 5, rowsGT)];
				stleft = resMat[rowres][2];//[index(rowres, 2, rowsST)];
				sttop = resMat[rowres][3];//[index(rowres, 3, rowsST)];
				stwidth = resMat[rowres][4];//[index(rowres, 4, rowsST)];
				stheight = resMat[rowres][5];//[index(rowres, 5, rowsST)];
				double dist = 1 - adjusted_boxiou(gtleft, gttop, gtwidth, gtheight, stleft, sttop, stwidth, stheight);
				if (dist <= threshold) 
				{
					alldist[i][j] = dist;
					// Add unique identifier to break ties
					alldist[i][j] += 1e-9 * uid;
					uid++;
				}
			}
		}

		vector<int> Lmate, Rmate;
	    MinCostMatching(alldist, Lmate, Rmate);
        
		for (size_t k = 0; k < Lmate.size(); k++) {
			if (alldist[k][Lmate[k]] == INF) continue;
			M[t][unmappedGt[k]] = unmappedEs[Lmate[k]];
			if (VERBOSE) printf("%d: map %d with %d\n", t+1, unmappedGt[k]+1, unmappedEs[Lmate[k]]+1);
		}

		vector<int> curtracked, alltrackers, mappedtrackers, falsepositives, set1;

		for (unordered_map<int,int>::iterator it = M[t].begin(); it != M[t].end(); it++) {
			curtracked.push_back(it->first);
			set1.push_back(it->second);
		}
		for (unordered_map<int,int>::iterator it = stInd[t].begin(); it != stInd[t].end(); it++) alltrackers.push_back(it->first);

		sort(set1.begin(), set1.end());
		sort(alltrackers.begin(), alltrackers.end());
		set_intersection(set1.begin(), set1.end(), alltrackers.begin(), alltrackers.end(), inserter(mappedtrackers, mappedtrackers.begin()));
		set_difference(alltrackers.begin(), alltrackers.end(), mappedtrackers.begin(), mappedtrackers.end(), inserter(falsepositives, falsepositives.end()));

		for (size_t k = 0; k < falsepositives.size(); k++) allfalsepos[t][falsepositives[k]] = falsepositives[k];

		//  mismatch errors
		if (t > 0)
		{
			for (size_t k = 0; k < curtracked.size(); k++)
			{
				int ct = curtracked[k];
				int lastnonempty = -1;
				for (int j = t - 1; j >= 0; j--) {
					if (M[j].find(ct) != M[j].end()) {
						lastnonempty = j; break;
					}
				}
				if (gtInd[t-1].find(ct)!=gtInd[t-1].end() && lastnonempty != -1)
				{
					int mtct = -1, mlastnonemptyct = -1;
					if (M[t].find(ct) != M[t].end()) mtct = M[t][ct];
					if (M[lastnonempty].find(ct) != M[lastnonempty].end()) mlastnonemptyct = M[lastnonempty][ct];

					if (mtct != mlastnonemptyct)
						mme[t]++;
				}
			}
		}

		c[t] = curtracked.size();
		for (size_t k = 0; k < curtracked.size(); k++)
		{
			int ct = curtracked[k];
			int eid = M[t][ct];
			
			double gtleft, gttop, gtwidth, gtheight, stleft, sttop, stwidth, stheight;
			int rowgt = gtInd[t][ct];
			int rowres = stInd[t][eid];
			gtleft = gtMat[rowgt][2];
			gttop = gtMat[rowgt][3];
			gtwidth = gtMat[rowgt][4];
			gtheight = gtMat[rowgt][5];
			stleft = resMat[rowres][2];
			sttop = resMat[rowres][3];
			stwidth = resMat[rowres][4];
			stheight = resMat[rowres][5];
			d[t][ct] = 1 - adjusted_boxiou(gtleft, gttop, gtwidth, gtheight, stleft, sttop, stwidth, stheight);
		}

		for (unordered_map<int,int>::iterator it = stInd[t].begin(); it != stInd[t].end(); it++) fp[t]++;
		fp[t] -= c[t];
		m[t] = g[t] - c[t];
	}
	
	// NOTE alltracked is M
	// partial_tracks: index 0 is mostly tracked(MT), 1 is partially tracked (PT), 2 is Mostly Losted (ML) 
	VI partial_tracks(3);

	for (int i = 0; i < Fgt; i++) {
		for (unordered_map<int, int>::iterator it = M[i].begin(); it != M[i].end(); it++) {
			int j = it->first;
			++M[i][j]; 
		}
	}
	calcPartialMisses(M, framesPerGT, partial_tracks);

	res.total_miss = sum(m); 
	double total_miss = res.total_miss;

	res.total_fp = sum(fp);
	double total_fp = res.total_fp;

	res.id_switches = sum(mme);
	double id_switches = res.id_switches;

	res.total_matches = sum(c);
	double total_matches = res.total_matches;

	res.unique_gt = (double)sum(unique_gt);
	
	res.mostly_tracked = partial_tracks[0];
	res.partially_tracked = partial_tracks[1];
	res.mostly_lossed = partial_tracks[2];

	res.total_frames = (double)Fgt;

	double num_gt = ((double)sum(g));
	res.total_gt = num_gt;

	res.distance = sum(d);

	// calculate metrics according to MOT16 paper
	res.MOTP = (1-sum(d)/total_matches) * 100;
	res.MOTAL = (1-((total_miss+total_fp+log10(id_switches+1))/num_gt))*100;
	res.MOTA = (1 - ((total_miss + total_fp + id_switches) / num_gt))*100;
	res.recall = total_matches/num_gt*100;
	res.precision = total_matches/(total_fp+total_matches)*100;
	res.FAR = total_fp/((double)Fgt);
}


// This function is a port of the MT, PT, ML calcutaion within CLEAR_MOT_HUN.m of the MOTdev-kit
void calcPartialMisses(vector<unordered_map<int,int>>& alltracked, vector<set<int>>& framesPerGT, VI& result) {

	for(size_t gt_id = 1; gt_id < framesPerGT.size(); ++gt_id){
		
		int trlengtha = 0;
		int total_gt = framesPerGT[gt_id].size();
		if(total_gt == 0)
			continue;

		// framesPerGT is based on gt ids (so starts at 1), but alltracked is zero based. Thus -1 within alltracked
		for(int frame : framesPerGT[gt_id])
			if(alltracked[frame -1][gt_id - 1] > 0)
				++trlengtha;

		double percent_tracked = (trlengtha / (double)total_gt);
		// Mostly tracked
		if( percent_tracked > 0.8)
			result[0]++;
		// Mostly lossed
		else if (percent_tracked < 0.2)
			result[2]++;
		// Partially tracked
		else  
			result[1]++;
	}
}


void computeTrackResults(vector<trackingResults>& results) {

	trackingResults all;

	for(auto & res : results) {

		all.IDTP += res.IDTP;
		all.IDFP += res.IDFP;
		all.IDFN += res.IDFN;

		all.total_tracks += res.total_tracks;
		all.total_gt += res.total_gt;
		all.total_frames += res.total_frames;
		all.distance += res.distance;

		all.total_matches += res.total_matches;
		all.total_miss += res.total_miss; 
		all.total_fp += res.total_fp;
		all.id_switches += res.id_switches;
			
		all.unique_gt += res.unique_gt;
		all.partially_tracked += res.partially_tracked;
		all.mostly_tracked += res.mostly_tracked;
		all.mostly_lossed += res.mostly_lossed;
	}

	all.MOTP = (1-all.distance/all.total_matches) * 100;
	all.MOTAL = (1-((all.total_miss+all.total_fp+log10(all.id_switches+1))/all.total_gt))*100;
	all.MOTA = (1 - ((all.total_miss + all.total_fp + all.id_switches) / all.total_gt))*100;
	all.recall = all.total_matches/all.total_gt *100;
	all.precision = all.total_matches/(all.total_fp+all.total_matches)*100;
	// if isnan(precision), precision=0; end % force to 0 if no matches found
	all.FAR = all.total_fp/all.total_frames;

	all.IDPrecision = (all.IDTP / (all.IDTP + all.IDFP) * 100);
	all.IDRecall = (all.IDTP / (all.IDTP + all.IDFN) * 100);
	all.IDF1 = (2*all.IDTP/(all.total_gt + all.total_tracks) * 100);

	string name = "All";
	printTrackResults(all, name);
}

void printTrackResults(trackingResults& res, string& seq_name) {

	cout << "Tracking quality results (higher is better, 100 max) for sequence: " << seq_name << endl << endl;

	//cout << "\tIDF1: " << res.IDF1 << endl;	
	/*cout << "\tIDPrecision: " << res.IDPrecision << endl;
	cout << "\tIDRecall: " << res.IDRecall << endl << endl;*/
	
	/*cout << "\tRecall: " << res.recall << endl;
	cout << "\tPrecision: " << res.precision << endl;	*/
	/*cout << "\tFAR: " << res.FAR << endl << endl;

	cout << "\tGT: " << res.unique_gt << endl;
	cout << "\tMT: " << res.mostly_tracked << endl;*/
	/*cout << "\tPT: " << res.partially_tracked << endl;
	cout << "\tML: " <<res.mostly_lossed << endl << endl;
	
	cout << "\tFP: " << res.total_fp << endl;	
	cout << "\tFN: " << res.total_miss << endl;	
	cout << "\tIDs: " << res.id_switches << endl << endl;	*/
	
	cout << "\tTracking accuracy (vs FPs, missed tgts, ID switches): " << res.MOTA << endl;
	cout << "\tTracking precision (bounding box alignment vs ground-truth): " << res.MOTP << endl;
	//cout << "\tMOTAL: " << res.MOTAL << endl << endl;
}
