#ifndef _EVALTRACK_
#define _EVALTRACK_

#include <string>
#include <iostream>
#include <vector>
#include <math.h>
#include <algorithm>
#include <map>
#include <iterator>
#include <set>
#include <unordered_map>
#include <numeric>

#include "utils.hpp"
#include "trackUtils.hpp"

using std::unordered_map;
using std::set;

struct trackingResults {
	// missed=sum(m);
	double total_miss; //= sum(m);
	// falsepositives=sum(fp);
	double total_fp; //= sum(fp);
	double id_switches; // idswitches=sum(mme);
	double total_matches;// = sum(c);
	double unique_gt; // = sum(g);
	double total_tracks;
	double total_gt;
	double mostly_tracked;
	double partially_tracked;
	double mostly_lossed;
	double total_frames;
	double distance;

	double MOTP; // = (1-sum(d)/total_matches) * 100;
	double MOTAL; //= (1-((total_miss+total_fp+log10(total_mme+1))/num_gt))*100;
	double MOTA; // = (1-(total_miss+total_fp+(total_mme)/num_gt))*100;
	double recall; // = total_matches/num_gt*100;
	double precision; // = total_matches/(total_fp+total_matches)*100;
	double FAR; // = total_fp/Fgt;

	double IDFP;
	double IDFN;
	double IDTP;
	double IDPrecision;
	double IDRecall;
	double IDF1;
};


void evalTrack(string& seq, VVVD& pre_proc_tracks, trackingResults& results);
void clear_mot_hun(VVD& gtMat, VVD& resMat, trackingResults& res);
void ID_measures(VVD& rawGT, VVD& rawTrack, trackingResults& res);

// calculates which how many gt ids were mostly tracked, partialty tracked, or mostly lossed during tracking.
void calcPartialMisses(vector<unordered_map<int,int>>& alltracked, vector<set<int>>& framesPerGT, VI& result);
// combines each sequences tracking results to produce overall score.
void computeTrackResults(vector<trackingResults>& results);
void printTrackResults(trackingResults& res, string& seq_name);


#endif
