#ifndef _TRACKUTILS_
#define _TRACKUTILS_

#include <vector>
#include <algorithm>
#include <cmath>
#include "utils.hpp"

// ByID is a functor used to sort gt and tracking data by ascending ids, ties broken with ascending frames
struct ByID {
	bool operator() (VD& lhs, VD& rhs){
		if( (int)lhs[1] != (int)rhs[1])
			return lhs[1] < rhs[1];
		else 
			return lhs[0] < rhs[0];
	}
};

// Takes long lists of gt and track data, converts them to be indexed by id
//		Updates gt, pred. Invalidates gtRaw and predRaw
void formatByID(VVD& gtRaw, VVD& predRaw, VVVD& gt, VVVD& pred);
// converts data from being frame indexed to a long list of row data
//		Updates data. Invalidates rawData
void formatToLongList(VVVD& rawData, VVD& data);

// The following are adapted from utils/MinCostMex.cpp, within MOT-devkit
double MinCostMatching(const VVD &cost, VI &Lmate, VI &Rmate);
	// maps row indices to col indecis such that cost of distMat is minimized
void MinCostAssignment(VVD& distMatrix, VI& assignment);

// The following are adapted from utils/costBlockMex.cpp, wihin MOT-devkit
void correspondingFrames(VVD& tr1, VVD& tr2, int* loc);
void computeDistances(VVD& tr1, VVD& tr2, int numPoints1, int numPoints2, int* position, double* distance);
void computeCostBlock(VVD& tr1, VVD& tr2, double threshold, double& cost, double& fp, double& fn);


void calcCostBlock(VVVD& gtRaw, VVVD& predRaw, VVD& cost,	VVD& fp, VVD& fn);
// The following are from utils/clearMOTMex.cpp, within MOT-devkit
inline double euclidean(double x1, double y1, double x2, double y2);
double adjusted_boxiou(double l1, double t1, double w1, double h1, double l2, double t2, double w2, double h2);

#endif
