#include "utils.hpp"

#include <string>
#include <math.h>
#include <sstream>
#include <iomanip>


// loads detections intp seq_dets from MOT style detection file
void loadtxt(vector<vector<double> > &seq_dets, ifstream &fin) {
    string line;
    while (getline(fin, line)) {
        stringstream ss(line);
        vector<double> row;
        double i;
        while (ss >> i) {
            row.push_back(i);
            char comma;
            ss >> comma;
        }
        seq_dets.push_back(row);
    }
}

vector<string> insert_seq_path(string start, vector<string>& sequences, string end){
    vector<string> long_seqs;
    for (auto& seq : sequences) 
        long_seqs.emplace_back(start + seq + end);
    return long_seqs;
}

void load_gt(vector<string>& seqs, VVVD& gt) { 
    size_t frame_shift = 0;
    for (auto& path : seqs) {
        ifstream gt_file;
        gt_file.open(path);
        // loads next seq data into gt and increments frame_shift by new frames amount
        loadtxt_by_frame(gt, gt_file, frame_shift);
        gt_file.close();
    }
}


// invalidates rawGT
void cleanGT(VVVD& rawGT, VVVD& gt, double threshhold) {

    for(size_t frame = 0; frame < rawGT.size(); ++frame) {
        for(auto &row: rawGT[frame]) {
            if( ((int)row[7] == 1) &&       // needed for both
                    (row[8] > threshhold ) && // for det (now passed in)
                        ((int)row[6] != 0) ) // for tracking, can be both
                gt[frame].emplace_back(move(row));            
        }
    }
}

// loads detections intp seq_dets from MOT style detection file
// loads next seq data into gt and increments frame_shift by new frames amount
void loadtxt_by_frame(vector<vector<vector<double> >> &seq_dets, ifstream &fin, size_t& frame_shift) {
    string line;
    size_t max_frame = 1;

    while (getline(fin, line)) {
        stringstream ss(line);
        vector<double> row;
        double i;
        while (ss >> i) {
            row.emplace_back(i);
            char comma;
            ss >> comma;
        }

        if((size_t)row[0] > max_frame)
            max_frame = (size_t)row[0];
        
        seq_dets[(int)(row[0]-1) + frame_shift].emplace_back(row);
    }
    frame_shift += max_frame;
}

// gets all files in folder
// original windows support remains untested
void get_all_files_names_within_folder(string& folder, vector<string>& names) {
    #ifdef WINDOWS
        char search_path[1000];
        sprintf_s(search_path, 1000, "%s*.*", folder.c_str());
        WIN32_FIND_DATA fd;
        HANDLE hFind = ::FindFirstFile(search_path, &fd);
        if (hFind != INVALID_HANDLE_VALUE) {
            do {
                // read all (real) files in current folder
                // , delete '!' read other 2 default folder . and ..
                if (!(fd.dwFileAttributes & FILE_ATTRIBUTE_DIRECTORY)) {
                    names.push_back(fd.cFileName);
                }
            } while (::FindNextFile(hFind, &fd));
            ::FindClose(hFind);
        }
    #else
        DIR *dpdf;
        struct dirent *epdf;
        dpdf = opendir(folder.c_str());
        if (dpdf != NULL){
            while ( (epdf = readdir(dpdf)) ){
                if(strncmp(epdf->d_name, ".", 1)!=0 && strncmp(epdf->d_name, "..", 2)!=0){
                    string filename = "/" ;
                    filename += epdf->d_name;
                    names.emplace_back(filename);
                }
            }
        }
    #endif
}

double calcBoxWeights(vector<double>& gt, vector<double>& result){
    double gt_x = gt[2]; double gt_y = gt[3]; 
    double gt_w = gt[4]; double gt_h = gt[5];
    
    double res_x = result[2]; double res_y = result[3]; 
    double res_w = result[4]; double res_h = result[5];
    
    // if no box overlap, not candidate for match
    if ((gt_x + gt_w < res_x) ||
        (gt_x > res_x + res_w)||
        (gt_y + gt_h < res_y) ||
        (gt_y > res_y + res_h))
        return 0;
    // else calc box overlap. 
    return boxiou(gt_x, gt_y, gt_w, gt_h, res_x, res_y, res_w, res_h);  
}

double boxIntersect(double bboxleft1, double bboxright1, double bboxbottom1, double bboxup1, double bboxleft2, double bboxright2, double bboxbottom2, double bboxup2) {
/*  % Compute intersection area of two bouding boxes A,B
    % A=[bboxleft1 bboxbottom1 abs(bboxright1-bboxleft1) abs(bboxbottom1-bboxup1)];
    % B=[bboxleft2 bboxbottom2 abs(bboxright2-bboxleft2) abs(bboxbottom2-bboxup2)];
    % 
    % isect=rectint(A,B);   
*/      
    double isect = 0.0;

    double hor = max(0.0,min(bboxright1,bboxright2) - max(bboxleft1,bboxleft2));
    if (!hor)
        return 0.0;

    double ver= max(0.0,min(bboxbottom1,bboxbottom2) - max(bboxup1,bboxup2));
    if (!ver)
        return 0.0;

    isect = hor*ver;

    return isect;
}

double boxUnion(double bboxleft1, double bboxright1, double bboxbottom1, double bboxup1, double bboxleft2, double bboxright2, double bboxbottom2, double bboxup2) {
    /* % Compute union of two bounding boxes */
    double a1 = bboxright1 - bboxleft1;
    double b1 = bboxbottom1 - bboxup1;
    double a2 = bboxright2 - bboxleft2;
    double b2 = bboxbottom2 - bboxup2;
    
    double box_union = a1*b1 + a2*b2;

    double bisect = boxIntersect(bboxleft1, bboxright1, bboxbottom1, bboxup1, bboxleft2, bboxright2, bboxbottom2, bboxup2);
  
    return box_union - bisect;
}


double boxiou(double x1,double y1,double w1,double h1,double x2,double y2,double w2,double h2) {
/* % compute intersection over union of two bboxes  */

    double bisect = boxIntersect(x1,x1+w1,y1+h1,y1,x2,x2+w2,y2+h2,y2);
    double iou = 0.0;
    if (bisect <=0 )
        return 0.0;
    
    double bunion = boxUnion(x1,x1+w1,y1+h1,y1,x2,x2+w2,y2+h2,y2);
    assert(bunion>0);
    iou = bisect/bunion;
    return iou;
}

// modifies assigment to map det index to corresponding gt index, -1 if no match
void boxHungarian(vector<vector<double>>& gt, vector<vector<double>>& data, vector<int>& assignment) { /* previously, return CSRMatrix<double> */
    
    int INF = 1e9;
    VVI costMat(data.size(), VI(gt.size(), INF)); 
    
    for (size_t d = 0; d < data.size(); ++d) {
        for(size_t g = 0; g < gt.size(); ++g) {
            double val = 1 - calcBoxWeights(gt[g], data[d]);
            if (val <= VIS_THRESH) {
                costMat[d][g] = (int)(val*10000);
            }
            //  NOTE: hungarian uses int, but data is double. So multiplying by 10000 
            //        before int cast to exagerate difference in weights.
        }
    }

    Hungarian::Result res = Hungarian::Solve(costMat, Hungarian::MODE_MINIMIZE_COST);
    if(!res.success){
        cout << "Hungarian failed in frame " << gt[0][0] << endl;
        exit(1);
    }

    assignment.resize(data.size(), -1);
    for(size_t d = 0; d < data.size(); ++d)
        for(size_t g = 0; g < res.assignment[d].size(); ++g)
            if((costMat[d][g] < INF) // Hungarian can match an INF row with INF col, which is not a true match.
                && (res.assignment[d][g] == 1) // match indicated with 1. 
                && (g < gt.size())) // algo seems to pad non-square mats, thus could assign to the invalid, extra columns.
            {
                assignment[d] = g;
                break; // there can only be one assignment
            }
}

// MOT16 and later omits certain detections from evaluation
bool isDistactor(int class_id) {
    return ( class_id == 2) // person on vehicle
        || ( class_id == 7) // static person
        || ( class_id == 8) // distractor
        || ( class_id == 12); // reflection
}

// gt and data are [frame x numDetInFrame x MOT16_info row]
// NOTE: invalidates data vector by moving rows into processed instead of copying 
void cleanData(VVVD& gt, VVVD& data, VVVD& processed, double threshhold) {

    // check that total frame values match
    assert(gt.size() == data.size());
        
    for(size_t frame = 0; frame < gt.size(); ++frame) {

        if(data[frame].empty() || gt[frame].empty())
            continue;

        VI assignment;
        boxHungarian(gt[frame], data[frame], assignment);
    
        for (size_t d = 0; d < data[frame].size(); d++) {
            int g = assignment[d];
            if( (g != -1) && 
                (gt[frame][g][8] < threshhold 
                  || isDistactor((int)gt[frame][g][7])) ) 
            {
                continue;
            }
            // det row is good (at least definitly not bad), keep it
            processed[frame].emplace_back(move(data[frame][d]));
        }
    }
}
