/*
    Utility functions to assist in file I/O and evaluation calcualtions
*/

#ifndef _UTILS_
#define _UTILS_

#include <stdlib.h>
#include <iostream>
#include <dirent.h>
#include <vector>
#include <fstream>
#include <sstream>
#include <string.h>
#include <numeric> //accumulate

#include <algorithm> // sorting
#include <math.h>    //abs
#include <cassert>
#include "hungarian.hpp"

#define VIS_THRESH 0.49  // default minimum cover % to consider a detection
//#define eps 0.001  // used for double comparison


using std::vector;
using std::cout;
using std::string;
using std::endl;
using std::ifstream;
using std::getline;
using std::stringstream;
using std::min;
using std::max;

typedef vector<double> VD;
typedef vector<VD> VVD;
typedef vector<VVD> VVVD;
typedef vector<VVVD> VVVVD;
typedef vector<int> VI;
typedef vector<VI> VVI;

template<class T>
T sum(vector<T>& _vector) {
    return std::accumulate(_vector.begin(), _vector.end(), 0.0);
}

template<class T>
T sum(vector<vector<T>>& matrix) {
    T total{};
    for(vector<T>& _vector: matrix)
        total += sum(_vector);
    return total;
}


// Takes MOT2015 style detection from fin, then loads them by row into seq_data.
//
// MOT2015 format: [frame_num, identify_num, left_bound, top_bound, width, height, confidence_score, x_coor, y_coor, z_coor]
// Note: for 2d detections, identify_num, x_coor, y_coor, z_coor are all -1
void loadtxt(vector<vector<double> > &seq_dets, std::ifstream &fin);
// 
void loadtxt_by_frame(vector<vector<vector<double> >> &seq_dets, ifstream &fin, size_t& frame_shift);

void load_gt(vector<string>& seqs, VVVD& gt);
vector<string> insert_seq_path(string start, vector<string>& sequences, string end);

// Collects the names of all files within folder.
void get_all_files_names_within_folder(std::string& folder, std::vector<std::string>& names);

// Removes extra ground truth data points (reflections, distractions, etc) that were used to clean detections or tracks
void cleanGT(VVVD& rawGT, VVVD& gt, double threshhold);

// The following functions are used to calculate the amount of area overlap between one gt and detection
double boxiou(double x1,double y1,double w1,double h1,double x2,double y2,double w2,double h2);
double boxUnion(double bboxleft1, double bboxright1, double bboxbottom1, double bboxup1, double bboxleft2, double bboxright2, double bboxbottom2, double bboxup2);
double boxIntersect(double bboxleft1, double bboxright1, double bboxbottom1, double bboxup1, double bboxleft2, double bboxright2, double bboxbottom2, double bboxup2);
double calcBoxWeights(vector<double>& gt, vector<double>& det);

// Calculates the overlap amount between each gt and data point within a frame, then applies the hungarian on those weights to assign best data to gt mapping that maximizes overlap area. Assignment will map data indices to their best matching gt index, or -1 for unmatched.
void boxHungarian(vector<vector<double>>& gt, vector<vector<double>>& data, vector<int>& assignment);

// MOT 16 and later use gound truth column 7 to label the type of object. Certain labels are considered distractors, and are ommited from detection and tracking evalution.
bool isDistactor(int class_id);

// Uses the Hungarian algorithm to match data points with gt, if any gt data is from the distractor list, then those data points are removed.
//  Also removes all data points associated with gt objects with visibilty below Threshold=0.5 for detection, 0 for tracking
//  All non-distractor data is moved to VVVD& processed. VVVD& data becomes invalidated.
void cleanData(VVVD& gt, VVVD& data, VVVD& processed, double threshhold);

#endif
